#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include "NebulaClient.h"
#include "IOTCClient.h"
#include "AVAPIs.h"
#include "P2PCam/AVFRAMEINFO.h"
#include "P2PCam/AVIOCTRLDEFs.h"
#include "LAN_WiFi_config/Client_LAN_WiFi_Config.h"
#include "disposable_params.h"
#include "getopt.h"
#include <windows.h>

#define ENABLE_DTLS 1 // 0:disable DTLS; 1:enable DTLS
#define ENABLE_RESEND 1 // 0:disable RESEND; 1:enable RESEND
#define AUDIO_BUF_SIZE 1024
#define VIDEO_BUF_SIZE 128000
#define AV_LIVE_STREAM_CHANNEL 0

#define PLAYBACK_PAUSE 0
#define PLAYBACK_START 1
#define PLAYBACK_STOP  2

#define AV_PLAYBACK_CHANNEL 1

enum BindType {
    DISABLE_BIND,
    SERVER_BIND,
    LOCAL_BIND
};

static void usleep(long long usec)
{
    HANDLE timer;
    LARGE_INTEGER interval;
    interval.QuadPart = -(10 * usec);

    timer = CreateWaitableTimer(NULL, TRUE, NULL);
    SetWaitableTimer(timer, &interval, 0, NULL, NULL, 0);
    WaitForSingleObject(timer, INFINITE);
    CloseHandle(timer);
}

static int gettimeofday(struct timeval *t, void *tz) {
    time_t _clock = 0;
    struct tm _tm = {0};
    SYSTEMTIME win_tm = {0};
    GetLocalTime(&win_tm);
    _tm.tm_year = win_tm.wYear - 1900;
    _tm.tm_mon = win_tm.wMonth - 1;
    _tm.tm_mday = win_tm.wDay;
    _tm.tm_hour = win_tm.wHour;
    _tm.tm_min = win_tm.wMinute;
    _tm.tm_sec = win_tm.wSecond;
    _tm.tm_isdst = -1;
    _clock = mktime(&_tm);
    t->tv_sec = _clock;
    t->tv_usec = win_tm.wMilliseconds * 1000;
    return 0;
}

//########################################################
//# Nebula configurations data
//########################################################
typedef struct CONN_INFO
{
    char license_key[512];
    char udid[MAX_PUBLIC_UDID_LENGTH+1];
    char pin_code[MAX_PIN_CODE_LENGTH+1];
    unsigned int connect_abort;
    unsigned int bind_abort;
    unsigned int send_command_abort;
    bool enable_playback;
    bool enable_speaker;
    bool enable_wakeup;
    bool enable_vsaas;
    unsigned int bind_type;
    char rent_am_token[MAX_TOKEN_LENGTH+1];
    char realm[MAX_REALM_LENGTH+1];
    char credential[512];
    int sid;
    int av_index;
    int speaker_time_sec;
    char vsaas_url[128];
    char vsaas_token[256];
}CONN_INFO;

static CONN_INFO gConnectionInfo = {0};

//########################################################
//# Get Timestamp
//########################################################
static unsigned int GetTimeStampMs()
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return (tv.tv_sec * 1000 + tv.tv_usec / 1000);
}

static void InitConnectionInfo(CONN_INFO *conn_info)
{
    memset(conn_info, 0, sizeof(CONN_INFO));
    conn_info->sid = -1;
    conn_info->av_index = -1;
    strncpy(conn_info->realm, "hausetopia-us", sizeof(conn_info->realm));
    conn_info->bind_type = SERVER_BIND;
}

//########################################################
//# Print error message
//########################################################
static void PrintErrHandling (int error)
{
    switch (error)
    {
    case IOTC_ER_SERVER_NOT_RESPONSE :
        //-1 IOTC_ER_SERVER_NOT_RESPONSE
        printf ("[Error code : %d]\n", IOTC_ER_SERVER_NOT_RESPONSE );
        printf ("Master doesn't respond.\n");
        printf ("Please check the network wheather it could connect to the Internet.\n");
        break;
    case IOTC_ER_FAIL_RESOLVE_HOSTNAME :
        //-2 IOTC_ER_FAIL_RESOLVE_HOSTNAME
        printf ("[Error code : %d]\n", IOTC_ER_FAIL_RESOLVE_HOSTNAME);
        printf ("Can't resolve hostname.\n");
        break;
    case IOTC_ER_ALREADY_INITIALIZED :
        //-3 IOTC_ER_ALREADY_INITIALIZED
        printf ("[Error code : %d]\n", IOTC_ER_ALREADY_INITIALIZED);
        printf ("Already initialized.\n");
        break;
    case IOTC_ER_FAIL_CREATE_MUTEX :
        //-4 IOTC_ER_FAIL_CREATE_MUTEX
        printf ("[Error code : %d]\n", IOTC_ER_FAIL_CREATE_MUTEX);
        printf ("Can't create mutex.\n");
        break;
    case IOTC_ER_FAIL_CREATE_THREAD :
        //-5 IOTC_ER_FAIL_CREATE_THREAD
        printf ("[Error code : %d]\n", IOTC_ER_FAIL_CREATE_THREAD);
        printf ("Can't create thread.\n");
        break;
    case IOTC_ER_FAIL_CREATE_SOCKET :
        //-6 IOTC_ER_FAIL_CREATE_SOCKET
        printf ("[Error code : %d]\n", IOTC_ER_FAIL_CREATE_SOCKET);
        printf ("Fails to create sockets.\n");
        break;
    case IOTC_ER_FAIL_SOCKET_OPT :
        //-7 IOTC_ER_FAIL_SOCKET_OPT
        printf ("[Error code : %d]\n", IOTC_ER_FAIL_SOCKET_OPT);
        printf ("Fails to set up socket options.\n");
        break;
    case IOTC_ER_FAIL_SOCKET_BIND :
        //-8 IOTC_ER_FAIL_SOCKET_BIND
        printf ("[Error code : %d]\n", IOTC_ER_FAIL_SOCKET_BIND);
        printf ("Fails to bind sockets.\n");
        break;
    case IOTC_ER_UNLICENSE :
        //-10 IOTC_ER_UNLICENSE
        printf ("[Error code : %d]\n", IOTC_ER_UNLICENSE);
        printf ("This UID is unlicense.\n");
        printf ("Check your UID.\n");
        break;
    case IOTC_ER_NOT_INITIALIZED :
        //-12 IOTC_ER_NOT_INITIALIZED
        printf ("[Error code : %d]\n", IOTC_ER_NOT_INITIALIZED);
        printf ("Please initialize the IOTCAPI first.\n");
        break;
    case IOTC_ER_TIMEOUT :
        //-13 IOTC_ER_TIMEOUT
        break;
    case IOTC_ER_INVALID_SID :
        //-14 IOTC_ER_INVALID_SID
        printf ("[Error code : %d]\n", IOTC_ER_INVALID_SID);
        printf ("This SID is invalid.\n");
        printf ("Please check it again.\n");
        break;
    case IOTC_ER_FAIL_GET_LOCAL_IP :
        //-16 IOTC_ER_FAIL_GET_LOCAL_IP
        printf ("[Error code : %d]\n", IOTC_ER_FAIL_GET_LOCAL_IP);
        printf ("Fails to get the client's local IP address.\n");
        break;
    case IOTC_ER_EXCEED_MAX_SESSION :
        //-18 IOTC_ER_EXCEED_MAX_SESSION
        printf ("[Error code : %d]\n", IOTC_ER_EXCEED_MAX_SESSION);
        printf ("[Warning]\n");
        printf ("The amount of session reach to the maximum.\n");
        printf ("It cannot be connected unless the session is released.\n");
        break;
    case IOTC_ER_CAN_NOT_FIND_DEVICE :
        //-19 IOTC_ER_CAN_NOT_FIND_DEVICE
        printf ("[Error code : %d]\n", IOTC_ER_CAN_NOT_FIND_DEVICE);
        printf ("Device didn't register on server, so we can't find device.\n");
        printf ("Please check the device again.\n");
        printf ("Retry...\n");
        break;
    case IOTC_ER_SESSION_CLOSE_BY_REMOTE :
        //-22 IOTC_ER_SESSION_CLOSE_BY_REMOTE
        printf ("[Error code : %d]\n", IOTC_ER_SESSION_CLOSE_BY_REMOTE);
        printf ("Session is closed by remote so we can't access.\n");
        printf ("Please close it or establish session again.\n");
        break;
    case IOTC_ER_REMOTE_TIMEOUT_DISCONNECT :
        //-23 IOTC_ER_REMOTE_TIMEOUT_DISCONNECT
        printf ("[Error code : %d]\n", IOTC_ER_REMOTE_TIMEOUT_DISCONNECT);
        printf ("We can't receive an acknowledgement character within a TIMEOUT.\n");
        printf ("It might that the session is disconnected by remote.\n");
        printf ("Please check the network wheather it is busy or not.\n");
        printf ("And check the device and user equipment work well.\n");
        break;
    case IOTC_ER_DEVICE_NOT_LISTENING :
        //-24 IOTC_ER_DEVICE_NOT_LISTENING
        printf ("[Error code : %d]\n", IOTC_ER_DEVICE_NOT_LISTENING);
        printf ("Device doesn't listen or the sessions of device reach to maximum.\n");
        printf ("Please release the session and check the device wheather it listen or not.\n");
        break;
    case IOTC_ER_CH_NOT_ON :
        //-26 IOTC_ER_CH_NOT_ON
        printf ("[Error code : %d]\n", IOTC_ER_CH_NOT_ON);
        printf ("Channel isn't on.\n");
        printf ("Please open it by IOTC_Session_Channel_ON() or IOTC_Session_Get_Free_Channel()\n");
        printf ("Retry...\n");
        break;
    case IOTC_ER_FAIL_CONNECT_SEARCH:
        //-27 IOTC_ER_FAIL_CONNECT_SEARCH
        printf ("[Error code : %d]\n", IOTC_ER_FAIL_CONNECT_SEARCH);
        printf ("The client stop connecting to the device.\n");
        break;
    case IOTC_ER_SESSION_NO_FREE_CHANNEL :
        //-31 IOTC_ER_SESSION_NO_FREE_CHANNEL
        printf ("[Error code : %d]\n", IOTC_ER_SESSION_NO_FREE_CHANNEL);
        printf ("All channels are occupied.\n");
        printf ("Please release some channel.\n");
        break;
    case IOTC_ER_TCP_TRAVEL_FAILED :
        //-32 IOTC_ER_TCP_TRAVEL_FAILED
        printf ("[Error code : %d]\n", IOTC_ER_TCP_TRAVEL_FAILED);
        printf ("Device can't connect to Master.\n");
        printf ("Don't let device use proxy.\n");
        printf ("Close firewall of device.\n");
        printf ("Or open device's TCP port 80, 443, 8080, 8000, 21047.\n");
        break;
    case IOTC_ER_TCP_CONNECT_TO_SERVER_FAILED :
        //-33 IOTC_ER_TCP_CONNECT_TO_SERVER_FAILED
        printf ("[Error code : %d]\n", IOTC_ER_TCP_CONNECT_TO_SERVER_FAILED);
        printf ("Device can't connect to server by TCP.\n");
        printf ("Don't let server use proxy.\n");
        printf ("Close firewall of server.\n");
        printf ("Or open server's TCP port 80, 443, 8080, 8000, 21047.\n");
        printf ("Retry...\n");
        break;
    case IOTC_ER_NO_PERMISSION :
        //-40 IOTC_ER_NO_PERMISSION
        printf ("[Error code : %d]\n", IOTC_ER_NO_PERMISSION);
        printf ("This UID's license doesn't support TCP.\n");
        break;
    case IOTC_ER_NETWORK_UNREACHABLE :
        //-41 IOTC_ER_NETWORK_UNREACHABLE
        printf ("[Error code : %d]\n", IOTC_ER_NETWORK_UNREACHABLE);
        printf ("Network is unreachable.\n");
        printf ("Please check your network.\n");
        printf ("Retry...\n");
        break;
    case IOTC_ER_FAIL_SETUP_RELAY :
        //-42 IOTC_ER_FAIL_SETUP_RELAY
        printf ("[Error code : %d]\n", IOTC_ER_FAIL_SETUP_RELAY);
        printf ("Client can't connect to a device via Lan, P2P, and Relay mode\n");
        break;
    case IOTC_ER_NOT_SUPPORT_RELAY :
        //-43 IOTC_ER_NOT_SUPPORT_RELAY
        printf ("[Error code : %d]\n", IOTC_ER_NOT_SUPPORT_RELAY);
        printf ("Server doesn't support UDP relay mode.\n");
        printf ("So client can't use UDP relay to connect to a device.\n");
        break;
    case IOTC_ER_NO_SERVER_LIST :
        //-44 IOTC_ER_NO_SERVER_LIST
        printf ("[Error code : %d]\n", IOTC_ER_NO_SERVER_LIST);
        printf ("No IOTC server information while client connect.\n");
        break;
    case IOTC_ER_DEVICE_MULTI_LOGIN :
        //-45 IOTC_ER_DEVICE_MULTI_LOGIN
        printf ("[Error code : %d]\n", IOTC_ER_DEVICE_MULTI_LOGIN);
        printf ("The connecting device duplicated loggin and may unconnectable.\n");
        break;
    case IOTC_ER_MASTER_NOT_RESPONSE :
        //-60 IOTC_ER_MASTER_NOT_RESPONSE
        printf ("[Error code : %d]\n", IOTC_ER_MASTER_NOT_RESPONSE);
        printf ("All masters have no respond.\n");
        break;
    case IOTC_ER_DEVICE_IS_SLEEP :
        //-64 IOTC_ER_DEVICE_IS_SLEEP
        printf ("[Error code : %d]\n", IOTC_ER_DEVICE_IS_SLEEP);
        printf ("Device is in sleep mode.\n");
        break;
    case IOTC_ER_DEVICE_REJECT_BY_WRONG_AUTH_KEY :
        // -68 IOTC_ER_DEVICE_REJECT_BY_WRONG_AUTH_KEY
        printf ("[Error code : %d]\n", IOTC_ER_DEVICE_REJECT_BY_WRONG_AUTH_KEY);
        printf ("Being rejected by device due to the wrong auth key.\n");
        break;
    case IOTC_ER_DEVICE_NOT_USE_KEY_AUTHENTICATION :
        // -69 IOTC_ER_DEVICE_NOT_USE_KEY_AUTHENTICATION
        printf ("[Error code : %d]\n", IOTC_ER_DEVICE_NOT_USE_KEY_AUTHENTICATION);
        printf ("Being rejected by device due to disable authentication.\n");
        break;

    default :
        break;
    }
}

//########################################################
//# Client receive video/audio from device
//########################################################
static void WINAPI ThreadReceiveAVData(void *arg)
{
    printf("[%s] Starting....\n", __func__);
    int av_index = *(int *)arg;
    int video_buf_size = VIDEO_BUF_SIZE;
    char *buf = (char*)malloc(video_buf_size);
    int ret;

    FRAMEINFO_t frame_info;
    char *video_format[] = {"MPEG4", "H263", "H264", "MJPEG", "UNKNOWN"};
    char *audio_format[] = {"ADPCM", "PCM", "SPEEX", "MP3", "G726", "UNKNOWN"};
    int codec_idx = 0;
    unsigned int frm_no;
    struct timeval tv, tv2;

    int v_fps = 0, a_fps = 0, round = 0, v_lost_cnt = 0, a_lost_cnt = 0;
    int out_buf_size = 0;
    int out_frm_size = 0;
    int out_frm_info_size = 0;
    int v_bps = 0, a_bps = 0;
    int video_not_ready = 1, audio_not_ready = 1;
    gettimeofday(&tv, NULL);

    while(1)
    {
        ret = avRecvFrameData2(av_index, buf, video_buf_size, &out_buf_size, &out_frm_size, (char *)&frame_info, sizeof(frame_info), &out_frm_info_size, &frm_no);
        // show Frame Info at 1st frame
        if(ret > 0 && frm_no == 0) {
            if(frame_info.codec_id == MEDIA_CODEC_VIDEO_MPEG4)
                codec_idx = 0;
            else if(frame_info.codec_id == MEDIA_CODEC_VIDEO_H263)
                codec_idx = 1;
            else if(frame_info.codec_id == MEDIA_CODEC_VIDEO_H264)
                codec_idx = 2;
            else if(frame_info.codec_id == MEDIA_CODEC_VIDEO_MJPEG)
                codec_idx = 3;
            else
                codec_idx = 4;

            printf("--- Recv Video Format: %s ---\n", video_format[codec_idx]);
        }

        if(ret == AV_ER_DATA_NOREADY) {
            //printf("Recv video, AV_ER_DATA_NOREADY[%d]\n", avIndex);
            //usleep(5 * 1000);
            video_not_ready = 1;
        } else if(ret == AV_ER_LOSED_THIS_FRAME) {
            printf("Recv video, Lost video frame NO[%d]\n", frm_no);
            v_lost_cnt++;
        } else if(ret == AV_ER_INCOMPLETE_FRAME) {
            #if 1
            if(out_frm_info_size > 0)
            printf("Incomplete video frame NO[%d] ReadSize[%d] FrmSize[%d] FrmInfoSize[%u] Codec[%d] Flag[%d]\n", frm_no, out_buf_size, out_frm_size, out_frm_info_size, frame_info.codec_id, frame_info.flags);
            else
            printf("Incomplete video frame NO[%d] ReadSize[%d] FrmSize[%d] FrmInfoSize[%u]\n", frm_no, out_buf_size, out_frm_size, out_frm_info_size);
            #endif
            v_lost_cnt++;
        } else if(ret == AV_ER_SESSION_CLOSE_BY_REMOTE) {
            printf("Recv video, AV_ER_SESSION_CLOSE_BY_REMOTE\n");
            break;
        } else if(ret == AV_ER_REMOTE_TIMEOUT_DISCONNECT) {
            printf("Recv video, AV_ER_REMOTE_TIMEOUT_DISCONNECT\n");
            break;
        } else if(ret == IOTC_ER_INVALID_SID) {
            printf("Recv video, Session cant be used anymore\n");
            break;
        } else if(ret == AV_ER_BUFPARA_MAXSIZE_INSUFF) {
            printf("Recv video, video buffer too small\n");
            video_buf_size = video_buf_size+(5*1024);
            buf = (char*)realloc(buf, video_buf_size);
            printf("increase buffer size to %d\n", video_buf_size);
            continue;
        } else {
            v_bps += out_buf_size;
            v_fps++;
            video_not_ready = 0;
            if(frame_info.tags == 1){
                printf("Receive the last video frame\n");
                break;
            }
        }

        ret = avRecvAudioData(av_index, buf, AUDIO_BUF_SIZE, (char *)&frame_info, sizeof(frame_info), &frm_no);
        // show Frame Info at 1st frame
        if(ret > 0 && frm_no == 0)
        {
            int idx = 0;
            if(frame_info.codec_id == MEDIA_CODEC_AUDIO_ADPCM)
                idx = 0;
            else if(frame_info.codec_id == MEDIA_CODEC_AUDIO_PCM)
                idx = 1;
            else if(frame_info.codec_id == MEDIA_CODEC_AUDIO_SPEEX)
                idx = 2;
            else if(frame_info.codec_id == MEDIA_CODEC_AUDIO_MP3)
                idx = 3;
            else if(frame_info.codec_id == MEDIA_CODEC_AUDIO_G726)
                idx = 4;
            else
                idx = 5;
            printf("--- Recv Audio Format: %s ---\n", audio_format[idx]);
        }

        if(ret == AV_ER_SESSION_CLOSE_BY_REMOTE) {
            printf("Recv audio, AV_ER_SESSION_CLOSE_BY_REMOTE\n");
            break;
        } else if(ret == AV_ER_REMOTE_TIMEOUT_DISCONNECT) {
            printf("Recv audio, AV_ER_REMOTE_TIMEOUT_DISCONNECT\n");
            break;
        } else if(ret == IOTC_ER_INVALID_SID) {
            printf("Recv audio, Session cant be used anymore\n");
            break;
        } else if(ret == AV_ER_LOSED_THIS_FRAME) {
            printf("Recv audio, AV_ER_LOSED_THIS_FRAME[%d]\n", frm_no);
            a_lost_cnt++;
        } else if (ret == AV_ER_DATA_NOREADY) {
            //printf("Recv audio, AV_ER_DATA_NOREADY[%d]\n", avIndex);
            //usleep(5 * 1000);
            audio_not_ready = 1;
        } else if(ret < 0) {
            printf("Recv audio, error[%d]!!!\n", ret);
            continue;
        } else {
            a_bps += ret;
            a_fps++;
            audio_not_ready = 0;
            if(frame_info.tags == 1){
                printf("Receive the last audio frame\n");
                break;
            }
        }

        gettimeofday(&tv2, NULL);
        long sec = tv2.tv_sec-tv.tv_sec, usec = tv2.tv_usec-tv.tv_usec;
        if(usec < 0) {
            sec--;
            usec += 1000000;
        }
        usec += (sec*1000000);

        if(usec > 1000000) {
            round++;
            printf("[AVAPIs_Client][avIndex:%d] Video FPS=%d, LostFrmCnt:%d, bps:%d Kbps, Audio FPS=%d, LostFrmCnt:%d, bps:%d Kbps\n",
                   av_index, v_fps, v_lost_cnt, (v_bps/1024)*8, a_fps, a_lost_cnt, (a_bps/1024)*8);
            gettimeofday(&tv, NULL);
            v_bps = 0;
            a_bps = 0;
            v_fps = 0;
            a_fps = 0;
        }
        if(1 == video_not_ready && 1 == audio_not_ready)
            usleep(5*1000);
    }

    if(buf) {
        free(buf);
    }
    printf("[%s] thread exit, avIndex[%d]\n", __func__, av_index);

    return 0;
}

//########################################################
//# Client send audio to device
//########################################################
static int StartSpeaker(int av_index);
static int StopSpeaker(int av_index);
static void WINAPI ThreadSpeaker(void *arg)
{
    CONN_INFO *conn_info = (CONN_INFO *)arg;
    char buf[AUDIO_BUF_SIZE];
    int frame_rate = 25;
    int sleep_us = 1000000/frame_rate;
    int ret = 0, av_index = 0;
    unsigned int start_time_sec = 0;
    FILE *fp = fopen("audio/8k_16bit_mono.raw", "rb");
    if(fp == NULL)
    {
        printf("[%s] fopen audio file error!!!\n", __func__);
        return;
    }
    
    av_index = conn_info->av_index;
    
    ret = StartSpeaker(av_index);
    if(ret != 0){
        printf("[%s] StartSpeaker error[%d]\n", __func__, ret);
        fclose(fp);
        return;
    }

    FRAMEINFO_t frame_info;
    memset(&frame_info, 0, sizeof(frame_info));

    printf("[%s] Starting avIndex[%d] \n", __func__, av_index);

    frame_info.codec_id = MEDIA_CODEC_AUDIO_PCM;
    frame_info.flags = (AUDIO_SAMPLE_8K << 2) | (AUDIO_DATABITS_16 << 1) | AUDIO_CHANNEL_MONO;
    start_time_sec = GetTimeStampMs()/1000;

    while(1) {
        int size = fread(buf, 1, 640, fp);
        if(size <= 0) {
            printf("rewind audio\n");
            rewind(fp);
            continue;
        }

        ret = avSendAudioData(av_index, buf, 640, &frame_info, sizeof(frame_info));
        if(ret == AV_ER_SESSION_CLOSE_BY_REMOTE) {
            printf("[%s] AV_ER_SESSION_CLOSE_BY_REMOTE\n", __func__);
            break;
        } else if(ret == AV_ER_REMOTE_TIMEOUT_DISCONNECT) {
            printf("[%s] AV_ER_REMOTE_TIMEOUT_DISCONNECT\n", __func__);
            break;
        } else if(ret == IOTC_ER_INVALID_SID) {
            printf("[%s] Session cant be used anymore\n", __func__);
            break;
        } else if(ret < 0) {
            printf("[%s] avSendAudioData error[%d]\n", __func__, ret);
            break;
        } else {
            //printf("[%s] avSendAudioData[%d]\n", __func__, ret);
        }
        usleep(sleep_us);

        //*** Speaker thread stop condition if necessary ***
        unsigned int diff_time_sec = (GetTimeStampMs()/1000)-start_time_sec;
        if(diff_time_sec > conn_info->speaker_time_sec){
            printf("[%s] elapsed[%u secs]\n", __func__, diff_time_sec);
            break;
        }
    }

    fclose(fp);
    StopSpeaker(av_index);
    printf("[%s] exit, avIndex[%d]\n", __func__, av_index);

    return;
}


//########################################################
//# Send live stream IOCtrl CMD to device
//########################################################
static int StartIpcamStream(int av_index)
{
    int ret;
    TUTKJsonObject *response = NULL;
    const char *request = "{\"func\":\"startVideo\",\"args\":{\"value\":true}}";
    printf("send json request: %s\n", request);

    ret = avSendJSONCtrlRequest(av_index, request, &response, 10);

    if(ret == AV_ER_NoERROR) {
        if(response)
            printf("get json response: %s\n", TUTK_Json_Obj_To_String(response));
    } else {
        printf("avSendJSONCtrlRequest error (%d)\n", ret);
        return -1;
    }
    // must free response here
    avFreeJSONCtrlResponse(response);


    request = "{\"func\":\"startAudio\",\"args\":{\"value\":true}}";
    printf("send json request: %s\n", request);

    ret = avSendJSONCtrlRequest(av_index, request, &response, 10);

    if(ret == AV_ER_NoERROR) {
        if(response)
            printf("get json response: %s\n", TUTK_Json_Obj_To_String(response));
    } else {
        printf("avSendJSONCtrlRequest error (%d)\n", ret);
        return -1;
    }
    // must free response here
    avFreeJSONCtrlResponse(response);

    return 0;
}

//########################################################
//# Send playback IOCtrl CMD to device
//########################################################
static int StartPlaybackRecord(int av_index)
{
    int ret;
    TUTKJsonObject *response = NULL;
    const char func[] = "playbackControl";
    const char file_name[] = "frames"; // hard coded filename in this sample
    int playback_ctrl = PLAYBACK_START;
    char request[256] = {0};
    snprintf(request, sizeof(request), "{\"func\":\"%s\",\"args\":{\"ctrl\":%d, \"fileName\":\"%s\"}}", func, playback_ctrl, file_name);
    printf("send json request: %s\n", request);

    ret = avSendJSONCtrlRequest(av_index, request, &response, 10);
    if(ret == AV_ER_NoERROR) {
        printf("get json response: %s\n\n", TUTK_Json_Obj_To_String(response));
    } else {
        printf("fail to get json response (%d)\n", ret);
        return -1;
    }

    // must free response
    avFreeJSONCtrlResponse(response);

    return 0;
}

static int StopPlaybackRecord(int av_index)
{
    int ret;
    TUTKJsonObject *response = NULL;
    const char func[] = "playbackControl";
    const char file_name[] = "frames"; // hard coded filename in this sample
    int playback_ctrl = PLAYBACK_STOP;
    char request[256] = {0};
    snprintf(request, sizeof(request), "{\"func\":\"%s\",\"args\":{\"ctrl\":%d, \"fileName\":\"%s\"}}", func, playback_ctrl, file_name);
    printf("send json request: %s\n", request);

    ret = avSendJSONCtrlRequest(av_index, request, &response, 10);
    if(ret == AV_ER_NoERROR) {
        printf("get json response: %s\n\n", TUTK_Json_Obj_To_String(response));
    } else {
        printf("fail to get json response (%d)\n", ret);
        return -1;
    }

    // must free response
    avFreeJSONCtrlResponse(response);

    return 0;
}


//########################################################
//# Send speaker IOCtrl CMD to device
//########################################################
#define ENABLE_SPEAKER true
#define DISABLE_SPEAKER false
static int SpeakerControl(int av_index, bool enable)
{
    int ret;
    TUTKJsonObject *response = NULL;
    const char func[] = "startSpeaker";
    char request[256] = {0};
    snprintf(request, sizeof(request), "{\"func\":\"%s\",\"args\":{\"value\":%s}}", func, enable?"true":"false");
    printf("send json request: %s\n", request);

    ret = avSendJSONCtrlRequest(av_index, request, &response, 10);
    if(ret == AV_ER_NoERROR) {
        printf("get json response: %s\n\n", TUTK_Json_Obj_To_String(response));
    } else {
        printf("fail to get json response (%d)\n", ret);
        return -1;
    }

    // must free response
    avFreeJSONCtrlResponse(response);

    return 0;
}

static int StartSpeaker(int av_index)
{
    return SpeakerControl(av_index, ENABLE_SPEAKER);
}

static int StopSpeaker(int av_index)
{
    return SpeakerControl(av_index, DISABLE_SPEAKER);
}


//########################################################
//# Thread - Create AV Connection and Create Video Receive thread
//########################################################
static int StartAvClient(int sid, unsigned char chid, unsigned int timeout_sec) {
    AVClientStartInConfig av_client_in_config;
    AVClientStartOutConfig av_client_out_config;
    memset(&av_client_in_config, 0, sizeof(av_client_in_config));
    memset(&av_client_out_config, 0, sizeof(av_client_out_config));

    av_client_in_config.cb = sizeof(av_client_in_config);
    av_client_in_config.iotc_session_id = sid;
    av_client_in_config.iotc_channel_id = chid;
    av_client_in_config.timeout_sec = timeout_sec;
    av_client_in_config.resend = ENABLE_RESEND;

    if(ENABLE_DTLS) {
        av_client_in_config.security_mode = AV_SECURITY_DTLS;
    } else {
        av_client_in_config.security_mode = AV_SECURITY_SIMPLE;
    }
    av_client_in_config.account_or_identity = "";
    av_client_in_config.password_or_token = "";
    av_client_in_config.auth_type = AV_AUTH_NEBULA;
    av_client_out_config.cb = sizeof(av_client_out_config);

    int av_index = avClientStartEx(&av_client_in_config, &av_client_out_config);

    printf("avClientStartEx: sid=[%d] chid=[%d] avIndex[%d] resend[%d] two_way_streaming[%d]\n", sid, chid,
            av_index, av_client_out_config.resend, av_client_out_config.two_way_streaming);

    return av_index;
}

static void WINAPI ThreadConnect(void *arg)
{
    CONN_INFO *conn_info = (CONN_INFO *)arg;
    int ret, live_stream_avindex = -1;
    int playback_avindex = -1;
    int receive_data_avindex = -1;
    int sid = conn_info->sid;
    int chid = AV_LIVE_STREAM_CHANNEL, playback_channel = AV_PLAYBACK_CHANNEL;
    int timeout_sec = 30;

    printf("%s start SID[%d]\n", __func__, sid);

    live_stream_avindex = StartAvClient(sid, chid, timeout_sec);
    if(live_stream_avindex < 0) {
        printf("  [%s:%d] avClientStartEx failed avIndex[%d]\n", __func__, __LINE__, live_stream_avindex);
        return;
    }
    conn_info->av_index = live_stream_avindex;

    if(conn_info->enable_playback){
        // send playback io ctrl command througth live stream av channel
        if (StartPlaybackRecord(live_stream_avindex) == 0) {
            // create playback av channel
            playback_avindex = StartAvClient(sid, playback_channel, timeout_sec);
        }
        if (playback_avindex < 0) {
            goto THREAD_CONNECT_EXIT;
        }
        receive_data_avindex = playback_avindex;
    } else {
        // send live stream io ctrl command
        if (StartIpcamStream(live_stream_avindex) != 0) {
            goto THREAD_CONNECT_EXIT;
        }
        receive_data_avindex = live_stream_avindex;
    }

    if(conn_info->enable_speaker){
        HANDLE thread_speaker_id = 0;
        thread_speaker_id = CreateThread(NULL, 0, ThreadSpeaker, (void *)conn_info, 0, NULL);
        if (thread_speaker_id == 0) {
            printf("  [%s:%d] pthread_create thread_Speaker failed\n", __func__, __LINE__);
        } else {
            CloseHandle(thread_speaker_id);
        }
    }

    HANDLE thread_avdata_id = 0;
    thread_avdata_id = CreateThread(NULL, 0, ThreadReceiveAVData, (void *)&receive_data_avindex, 0, NULL);
    if (thread_avdata_id == 0) {
        printf("  [%s:%d] pthread_create thread_ReceiveAVData failed\n", __func__, __LINE__);
        goto THREAD_CONNECT_EXIT;
    }
    WaitForSingleObject(thread_avdata_id, INFINITE);
    CloseHandle(thread_avdata_id);
    

THREAD_CONNECT_EXIT:
    if(conn_info->enable_playback) {
        StopPlaybackRecord(live_stream_avindex);
        if(playback_avindex >= 0){
            printf("  [%s:%d] avClientStop playback avIndex[%d]\n", __func__, __LINE__, playback_avindex);
            avClientStop(playback_avindex);
        }
        conn_info->enable_playback = 0;
    }
    avClientStop(live_stream_avindex);
    printf("  [%s:%d] avClientStop avIndex[%d]\n", __func__, __LINE__, live_stream_avindex);
    IOTC_Session_Close(sid);
    printf("  [%s:%d] IOTC_Session_Close SID[%d]\n", __func__, __LINE__, sid);
    printf("%s: exit!! SID[%d]\n", __func__, sid);
    return;
}

//########################################################
//# Bind
//########################################################
static int ClientSeverBind(CONN_INFO *conn_info, NebulaClientCtx **client_ctx, unsigned int timeout_ms) {
    int ret = 0;
    TUTKJsonObject *json_args = NULL;
    const char *resp_str = NULL;
    const char *identity_value = NULL;
    const char *avtoken_value = NULL;

    ret = Nebula_Client_New(conn_info->udid, client_ctx);
    printf("Nebula_Client_New ret[%d]\n", ret);
    if (ret != NEBULA_ER_NoERROR) {
        return -1;
    }

    ret = Nebula_Client_Bind(*client_ctx, conn_info->pin_code, &json_args, timeout_ms, &conn_info->bind_abort);
    printf("Nebula_Client_Bind ret[%d]\n", ret);
    if (ret != NEBULA_ER_NoERROR) {
        return -1;
    }

    resp_str = TUTK_Json_Obj_To_String(json_args);
    printf("[Bind_Response]\n%s\n", resp_str);

    ret = TUTK_Json_Obj_Get_Sub_Obj_String(json_args, "identity", &identity_value);
    if (ret == NEBULA_ER_NoERROR) {
        printf("Get identity: %s\n", identity_value);
    }
    
    TUTK_Json_Obj_Get_Sub_Obj_String(json_args, "avtoken", &avtoken_value);
    if (ret == NEBULA_ER_NoERROR) {
        printf("Get avtoken: %s\n", avtoken_value);
    }

    ret = Nebula_Client_Free_Bind_Response(json_args);
    if (ret != NEBULA_ER_NoERROR) {
        printf("Nebula_Client_Free_Bind_Response failed!\n");
    }

    return 0;
}

static void PrintUsage()
{
    printf("#########################################################################\n");
    printf("./Nebula_Client_AV [options]\n");
    printf("[options]\n");
    printf("[M]\t-k [LicenseKey]\t\t\tTUTK license key.\n");
    printf("[M]\t-u [UDID]\t\t\tConfig UDID.\n");
    printf("[O]\t-r [Realm]\t\t\tThe realm for rent IOTC UID. If use static UDID, realm is not required.\n");
    printf("[O]\t-t [AM token]\t\t\tThe token for rent IOTC UID. If use static UDID, token is not required.\n");
    printf("[O]\t-b [l(ocal)|s(erver)|n(one)]\tBinding mode. Bind in local, bind through TUTK bind server or do not bind. Default is server.\n");
    printf("[O]\t-c [Credential]\t\t\tThe credential for create Nebula client context.\n");
    printf("[O]\t-s [Time]\t\t\tDemo speaker for time in seconds. Disabled by default.\n");
    printf("[O]\t-v [Info]\t\t\tSend VSaaS info to device, enable device to notify VSaaS server pull stream. Disabled by default.\n");
    printf("   \t   [Info]\t\t\tFormat:{\"url\": URL, \"stoken\": Token}, Example: {\"url\": \"https://vsaas.com/api/stream_url\", \"stoken\": \"QAZwsxEDCrfv\"}\n");
    printf("[O]\t-p \t\t\t\tDemo playback. Disabled by default.\n");
    printf("[O]\t-w \t\t\t\tDemo wakeup. Disabled by default.\n");
    printf("[O]\t-h \t\t\t\tShow usage\n");
}

static int ParseInputOptions(int argc, char *argv[], CONN_INFO *conn_info)
{
    int option = 0;
    while((option = getopt(argc, argv, "k:u:c:t:r:b:s:v:phw")) > 0){
        switch(option) {
          case 'k':
            memset(conn_info->license_key,0,512);
            strcpy(conn_info->license_key, optarg);
            break;
          case 'u':
            strncpy(conn_info->udid, optarg, sizeof(conn_info->udid));
            conn_info->udid[MAX_PUBLIC_UDID_LENGTH] = '\0';
            printf("UDID %s\n", conn_info->udid);
            break;
          case 'c':
            strncpy(conn_info->credential, optarg, sizeof(conn_info->credential));
            break;
          case 't':
            strncpy(conn_info->rent_am_token, optarg, sizeof(conn_info->rent_am_token));
            break;
          case 'r':
            strncpy(conn_info->realm, optarg, sizeof(conn_info->realm));
            break;
          case 'b':
            if((strlen(optarg)==strlen("l") && strncmp(optarg, "l", 1) == 0)
               || (strlen(optarg)==strlen("local") && strncmp(optarg, "local", strlen("local")) == 0)) {
                conn_info->bind_type = LOCAL_BIND;
            } else if((strlen(optarg)==strlen("s") && strncmp(optarg, "s", 1) == 0)
               || (strlen(optarg)==strlen("server") && strncmp(optarg, "server", strlen("server")) == 0)) {
                conn_info->bind_type = SERVER_BIND;
            } else if((strlen(optarg)==strlen("n") && strncmp(optarg, "n", 1) == 0)
               || (strlen(optarg)==strlen("none") && strncmp(optarg, "none", strlen("none")) == 0)) {
                conn_info->bind_type = DISABLE_BIND;
            } else {
                printf("Unknown optarg %s\n", optarg);
                return -1;
            }
            break;
          case 'p':
            conn_info->enable_playback = 1;
            break;
          case 's':
            conn_info->enable_speaker = 1;
            conn_info->speaker_time_sec = atoi(optarg);
            break;
          case 'v': {
            TUTKJsonObject *json_obj = NULL;
            int ret = TUTK_Json_Obj_Create_From_String(optarg, &json_obj);
            if(ret != NEBULA_ER_NoERROR){
                printf("Input vsaas info not json format\n");
                return -1;
            }
            const char *url = NULL, *token = NULL;
            ret = TUTK_Json_Obj_Get_Sub_Obj_String(json_obj, "url", &url);
            if(ret != NEBULA_ER_NoERROR){
                printf("Input vsaas info no url\n");
                TUTK_Json_Obj_Release(json_obj);
                return -1;
            }
            ret = TUTK_Json_Obj_Get_Sub_Obj_String(json_obj, "stoken", &token);
            if(ret != NEBULA_ER_NoERROR){
                printf("Input vsaas info no stoken\n");
                TUTK_Json_Obj_Release(json_obj);
                return -1;
            }
            strcpy(conn_info->vsaas_url, url);
            strcpy(conn_info->vsaas_token, token);
            conn_info->enable_vsaas = 1;
            TUTK_Json_Obj_Release(json_obj);
            }
            break;
          case 'w':
            conn_info->enable_wakeup = 1;
            break;
          case 'h':
            return -1;
          default:
            printf("Unknown option %c\n", option);
            return -1;
        }
    }
    if(strlen(conn_info->udid) == 0){
        printf("Must specify UDID\n");
        return -1;
    }
    return 0;
}

//########################################################
//# Main function
//########################################################
int main(int argc, char *argv[])
{
    int ret = 0;
    NebulaClientCtx *client_ctx = NULL;
    TUTKJsonObject *json_response = NULL;
    const char *profile_str = NULL;
    CONN_INFO *conn_info = &gConnectionInfo;

    InitConnectionInfo(conn_info);

    /* Check command line options */
    ret = ParseInputOptions(argc, argv, conn_info);
    if(ret < 0){
        printf("ParseInputOptions ret[%d]\n", ret);
        PrintUsage();
        return -1;
    }
    
    ret = loadDisposableParams(conn_info->pin_code, NULL, NULL);
    if(ret<0){
        printf("loadDisposableParams ret[%d], %s exit...!!\n", ret, argv[0]);
        return -1;
    }

    ret = TUTK_SDK_Set_License_Key(conn_info->license_key);
    printf("TUTK_SDK_Set_License_Key: %d\n", ret);
    if (ret < 0) {
        return -1;
    }

    //Set log file path
    LogAttr log_attr;
    memset(&log_attr, 0, sizeof(log_attr));
    char log_path[] = "./client_log.txt";
    log_attr.path = log_path;

    ret = IOTC_Set_Log_Attr(log_attr);
    printf("IOTC_Set_Log_Attr ret[%d]\n", ret);

    //Nebula Initialize
    ret = Nebula_Initialize();
    printf("Nebula_Initialize ret[%d]\n", ret);
    if (ret != NEBULA_ER_NoERROR) {
        printf("%s exit...!!\n", argv[0]);
        return -1;
    }

    // IOTC Initialize
    ret = IOTC_Initialize2(0);
    printf("IOTC_Initialize2 ret[%d]\n", ret);
    if (ret != IOTC_ER_NoERROR) {
        printf("%s exit...!!\n", argv[0]);
        PrintErrHandling(ret);
        return -1;
    }

    // AV Initialize
    ret = avInitialize(32);
    printf("avInitialize ret[%d]\n", ret);

    // wakeup demo ignore bind process
    if(conn_info->enable_wakeup == 0){
        switch (conn_info->bind_type) {
        case SERVER_BIND:
            printf("Bind type : Server Bind\n");
            ret = ClientSeverBind(conn_info, &client_ctx, 30000);
            printf("ClientSeverBind ret[%d]\n", ret);
            if (ret != NEBULA_ER_NoERROR) {
                printf("%s exit...!!\n", argv[0]);
                return -1;
            }
            break;

        case LOCAL_BIND:
            printf("Bind type : Local Bind\n");
            ret = Client_LAN_WIFI_Config(conn_info->udid, 0);
            printf("Client_LAN_WIFI_Config ret[%d]\n", ret);
            if (ret != 0) {
                printf("%s exit...!!\n", argv[0]);
                return -1;
            }
            Client_LAN_Get_Bind_Credential(conn_info->credential);
            if (strlen(conn_info->credential) == 0) {
                printf("get CTX string fail, %s exit...!!\n", argv[0]);
                return -1;
            }
            break;
        case DISABLE_BIND:
            printf("Bind type : Disable Bind\n");
            break;
        default:
            printf("Unknown Bind type[%d]\n", conn_info->bind_type);
            return -1;
            break;
        }
    }

    if (client_ctx == NULL) {
        if (strlen(conn_info->credential) == 0) {
            printf("Empty credential, %s exit...!!\n", argv[0]);
            return -1;
        }
        ret = Nebula_Client_New_From_String(conn_info->udid, conn_info->credential, &client_ctx);
        printf("Nebula_Client_New_From_String ret[%d]\n", ret);
        if (ret != NEBULA_ER_NoERROR) {
            printf("%s exit...!!\n", argv[0]);
            return -1;
        }
    }

    if(conn_info->enable_wakeup == 1){
        unsigned int wakeup_abort = 0;
        for (int i = 0; i < 5; i++) {
            ret = Nebula_Client_Wakeup_Device(client_ctx, 10000, &wakeup_abort);

            if (i != 0 && ret == NEBULA_ER_DEVICE_ONLINE) {
                printf("Device awakened.\n");
                break;
            } else if (ret != NEBULA_ER_NoERROR) {
                printf("Nebula_Client_Wakeup_Device fails: %d!!\n", ret);
                return -1;
            }
            Sleep(1000);
        }
    }

    char request1[] = "{\"func\":\"getProfile\"}";
    printf("get device's profile : \n%s\n", request1);
    ret = Nebula_Client_Send_Command(client_ctx, request1, &json_response, 10000, &conn_info->send_command_abort);
    printf("json_response [%d] \n\n", ret);
    if (ret != NEBULA_ER_NoERROR) {
        printf("%s exit...!!\n", argv[0]);
        return -1;
    }

    profile_str = TUTK_Json_Obj_To_String(json_response);
    printf("[getProfile]\n%s\n", profile_str);

    ret = Nebula_Client_Free_Send_Command_Response(json_response);
    if (ret != NEBULA_ER_NoERROR)
        printf("Nebula_Client_Free_Send_Command_Response failed!\n");

    char request2[] = "{\"func\":\"getNightMode\"}";
    printf("get device's night mode : \n%s\n", request2);
    ret = Nebula_Client_Send_Command(client_ctx, request2, &json_response, 10000, &conn_info->send_command_abort);
    printf("json_response [%d]\n\n", ret);
    if (ret != NEBULA_ER_NoERROR) {
        printf("%s exit...!!\n", argv[0]);
        return -1;
    }

    profile_str = TUTK_Json_Obj_To_String(json_response);
    printf("[getNightMode]\n%s\n", profile_str);

    ret = Nebula_Client_Free_Send_Command_Response(json_response);
    if (ret != NEBULA_ER_NoERROR)
        printf("Nebula_Client_Free_Send_Command_Response failed!\n");

    if(conn_info->enable_playback){
        //Get event list storaged in device
        const TUTKJsonObject *content_object = NULL;
        const TUTKJsonObject *value_object = NULL;
        const TUTKJsonObject *event_object = NULL;
        int event_cnt = 0;
        char event_request[] = "{\"func\":\"queryEventList\",\"args\":{\"startTime\":1559318400, \"endTime\":1560182400,\"eventType\":\"motion\"}}";
        printf("***Send command to query event list: ***\n%s\n", event_request);
        ret = Nebula_Client_Send_Command(client_ctx, event_request, &json_response, 10000, &conn_info->send_command_abort);
        if (ret != NEBULA_ER_NoERROR) {
            printf("%s exit...!!\n", argv[0]);
            return -1;
        }

        //show event list here
        ret = TUTK_Json_Obj_Get_Sub_Obj(json_response, "content", &content_object);
        ret = TUTK_Json_Obj_Get_Sub_Obj(content_object, "value", &value_object);
        event_cnt = TUTK_Json_Obj_Get_Array_length(value_object);
        printf("[queryEventList response]\n Event list count[%d]\n", event_cnt);
        for(int i = 0; i < event_cnt; i++){
            TUTK_Json_Obj_Get_Array_Element(value_object, i, &event_object);
            printf("/**********NO.%d*************/\n", i+1);
            printf("%s\n", TUTK_Json_Obj_To_String(event_object));
        }
        printf("/***************************/\n\n");


        ret = Nebula_Client_Free_Send_Command_Response(json_response);
        if (ret != NEBULA_ER_NoERROR)
            printf("Nebula_Client_Free_Send_Command_Response failed!\n");
    }
    
    if(conn_info->enable_vsaas) {
        char set_vsaas_info_cmd[512] = {0};
        snprintf(set_vsaas_info_cmd, sizeof(set_vsaas_info_cmd), "{\"func\": \"setCloudRecordingEndpoint\", \"args\": {\"url\": \"%s/%s?stoken=%s\"}}", conn_info->vsaas_url, conn_info->udid, conn_info->vsaas_token);
        printf("***Send command to enable vsaas: ***\n%s\n", set_vsaas_info_cmd);
        ret = Nebula_Client_Send_Command(client_ctx, set_vsaas_info_cmd, &json_response, 10000, &conn_info->send_command_abort);
        if (ret != NEBULA_ER_NoERROR) {
            printf("%s exit...!!\n", argv[0]);
            return -1;
        }

        if(json_response != NULL) {
            printf("vsaas resp [%s]\n", TUTK_Json_Obj_To_String(json_response));
        }

        ret = Nebula_Client_Free_Send_Command_Response(json_response);
        if (ret != NEBULA_ER_NoERROR)
            printf("Nebula_Client_Free_Send_Command_Response failed!\n");
    }

    ret = IOTC_Client_Connect_By_Nebula(client_ctx, conn_info->rent_am_token, conn_info->realm, 60000, &conn_info->connect_abort);

    printf("IOTC_Client_Connect_By_Nebula ret[%d]\n", ret);
    if (ret >= IOTC_ER_NoERROR) {
        printf("Session[%d] connect!\n", ret);
        struct st_SInfoEx session_info;
        session_info.size = sizeof(struct st_SInfoEx);
        if(IOTC_Session_Check_Ex(ret, &session_info) == IOTC_ER_NoERROR) {
            char *mode[3] = {"P2P", "RLY", "LAN"};
            if( isdigit( session_info.RemoteIP[0] )){
                printf("device is from[IP:%s, Port:%d] Mode[%s] VPG[%d:%d:%d] VER[%X] NAT[%d] AES[%d]\n", \
                session_info.RemoteIP, session_info.RemotePort, mode[(int)session_info.Mode], session_info.VID, session_info.PID, session_info.GID, session_info.IOTCVersion, session_info.RemoteNatType, session_info.isSecure);
            }
            conn_info->sid = ret;
            HANDLE connect_thread_id = 0;
            connect_thread_id = CreateThread(NULL, 0, ThreadConnect, (void *)conn_info, 0, NULL);
            if (connect_thread_id == 0) {
                printf("[%s:%d] pthread_create ThreadConnect failed ret[%d]\n", __func__, __LINE__, ret);
                avDeInitialize();
                IOTC_DeInitialize();
                return -1;
            }
            WaitForSingleObject(connect_thread_id, INFINITE);
            CloseHandle(connect_thread_id);
        }
    } else {
        PrintErrHandling(ret);
        printf("IOTC_Client_Connect_By_Nebula exit...!!\n");
    }

    Nebula_Client_Delete(client_ctx);

    avDeInitialize();
    printf("avDeInitialize\n");
    IOTC_DeInitialize();
    printf("IOTC_DeInitialize\n");
    Nebula_DeInitialize();
    printf("Nebula_DeInitialize\n");
    return 0;
}
